using SimpleJSON;
using System;
using System.Collections;
using UnityEngine;
using UnityEngine.EventSystems;

namespace AwasUI {

    internal class AwasUIButton {
        public EventTrigger buttonTrigger;
        public EventTrigger buttonAltTrigger;
        private string _buttonLabel;
        private string _buttonAltLabel;
        private Color _buttonColor;
        private float _buttonAlpha;
        private Color _textColor;
        private float _textAlpha;
        private bool _hidden = false;
        private readonly MVRScript script;
        public readonly string configUiName;
        public readonly float worldXpos;
        public readonly float worldYpos;
        public readonly float buttonWidth = 0;
        public readonly float buttonHeight = 0;
        public UIDynamicButton uiButton;
        private Action onButtonDownCallback;

        public string buttonLabel {
            get { return _buttonLabel; }
            set {
                _buttonLabel = value;
                OnUpdate();
            }
        }

        public string buttonAltLabel {
            get { return _buttonAltLabel; }
            set {
                _buttonAltLabel = value;
                OnUpdate();
            }
        }

        public Color buttonColor {
            get { return _buttonColor; }
            set {
                _buttonColor = value;
                _buttonColor.a = _buttonAlpha;
                OnUpdate();
            }
        }

        public float buttonAlpha {
            get { return _buttonAlpha; }
            set {
                _buttonAlpha = value;
                _buttonColor.a = value;
                OnUpdate();
            }
        }

        public Color textColor {
            get { return _textColor; }
            set {
                _textColor = value;
                _textColor.a = _textAlpha;
                OnUpdate();
            }
        }

        public float textAlpha {
            get { return _textAlpha; }
            set {
                _textAlpha = value;
                _textColor.a = value;
                OnUpdate();
            }
        }

        public bool hidden {
            get { return _hidden; }
            set {
                _hidden = value;
                OnUpdate();
            }
        }

        public AwasUIButton(MVRScript script, string configUiName, string buttonLabel, string buttonAltLabel, float worldXpos, float worldYpos) {
            this.script = script;
            this.configUiName = configUiName;
            string triggerName = "Trigger";
            string triggerAltName = "TriggerAlt";

            buttonTrigger = new EventTrigger(this.script, triggerName);
            buttonAltTrigger = new EventTrigger(this.script, triggerAltName);
            _buttonLabel = buttonLabel;
            _buttonAltLabel = buttonAltLabel;
            _buttonColor = new Color(0.23f, 0.10f, 0.18f) { a = _buttonAlpha = 0.25f };
            _textColor = new Color(0.91f, 0.78f, 0.85f) { a = _textAlpha = 0.8f };
            this.worldXpos = worldXpos;
            this.worldYpos = worldYpos;
        }

        public AwasUIButton(MVRScript script, string configUiName, string buttonLabel, string buttonAltLabel, float worldXpos, float worldYpos, float buttonWidth = 0, float buttonHeight = 0)
            : this(script, configUiName, buttonLabel, buttonAltLabel, worldXpos, worldYpos) {
            this.buttonWidth = buttonWidth;
            this.buttonHeight = buttonHeight;
        }

        public void ResetUiButtonToDefaultState() {
            if (uiButton != null) {
                uiButton.label = buttonLabel;
                uiButton.buttonColor = buttonColor;
                uiButton.buttonText.color = textColor;
                uiButton.transform.localScale = Vector3.one;
            }
        }

        private void OnUpdate() {
            if (uiButton != null) {
                uiButton.label = buttonLabel;
                if (hidden) {
                    uiButton.button.interactable = false;
                    uiButton.button.enabled = false;

                    uiButton.buttonColor = Color.clear;
                    uiButton.buttonText.color = Color.clear;
                } else {
                    uiButton.button.interactable = true;
                    uiButton.button.enabled = true;

                    uiButton.buttonColor = buttonColor;
                    uiButton.buttonText.color = textColor;
                }
            }
        }

        private float timeDown = 0;
        private bool buttonIsDown = false;
        private bool cancelButtonAction = false;
        private IEnumerator ButtonLongPressModifier(UIDynamicButton uiButton, string altLabel) {
            float l = 0;
            float startAlpha = uiButton.buttonColor.a;
            Color buttonColor = uiButton.buttonColor;
            float endAlpha = startAlpha <= 0.5 ? 1 : 0;
            float lSummand = 0.5f / 20f;
            while (Time.time - timeDown <= 0.5f && buttonIsDown) {
                buttonColor.a = Mathf.Lerp(startAlpha, endAlpha, l);
                uiButton.buttonColor = buttonColor;
                uiButton.transform.localScale = Vector3.Lerp(Vector3.one, Vector3.one * 1.1f, l);
                l += (1f / 20);
                yield return new WaitForSeconds(lSummand);
            }
            uiButton.transform.localScale = Vector3.one;
            if (buttonIsDown) {
                uiButton.label = altLabel;
            }
        }

        public void SetUiButton(UIDynamicButton button) {
            uiButton = button;
            UnityEngine.EventSystems.EventTrigger trigger = uiButton.gameObject.AddComponent<UnityEngine.EventSystems.EventTrigger>();

            var pointerDown = new UnityEngine.EventSystems.EventTrigger.Entry();
            pointerDown.eventID = EventTriggerType.PointerDown;
            pointerDown.callback.AddListener((e) => {
                cancelButtonAction = false;
                if (hidden) return;
                timeDown = Time.time;
                buttonIsDown = true;
                uiButton.StartCoroutine(ButtonLongPressModifier(uiButton, buttonAltLabel));
                if (onButtonDownCallback != null) onButtonDownCallback();
            });

            var pointerOut = new UnityEngine.EventSystems.EventTrigger.Entry();
            pointerOut.eventID = EventTriggerType.PointerExit;
            pointerOut.callback.AddListener((e) => {
                cancelButtonAction = true;
                if (buttonIsDown) ResetUiButtonToDefaultState();
            });

            var pointerUp = new UnityEngine.EventSystems.EventTrigger.Entry();
            pointerUp.eventID = EventTriggerType.PointerUp;
            pointerUp.callback.AddListener((e) => {
                if (hidden) return;
                ResetUiButtonToDefaultState();
                buttonIsDown = false;
                if (cancelButtonAction) return;
                if (Time.time - timeDown > 0.5f) {
                    buttonAltTrigger.Trigger();
                } else {
                    buttonTrigger.Trigger();
                    uiButton.StopCoroutine("ButtonColorator");
                }
            });

            trigger.triggers.Add(pointerDown);
            trigger.triggers.Add(pointerUp);
            trigger.triggers.Add(pointerOut);
        }

        public void CancelButtonAction() {
            cancelButtonAction = true;
        }

        public bool IsButtonActionActive() {
            return buttonIsDown;
        }

        public void OnButtonDown(Action buttonDownCallback) {
            onButtonDownCallback = buttonDownCallback;
        }

        public JSONClass GetJSON() {
            JSONClass jsonClass = new JSONClass {
                ["configUiName"] = configUiName,
                [buttonTrigger.Name] = buttonTrigger.GetJSON(script.subScenePrefix),
                [buttonAltTrigger.Name] = buttonAltTrigger.GetJSON(script.subScenePrefix),
                ["buttonLabel"] = _buttonLabel,
                ["buttonAltLabel"] = _buttonAltLabel,
                ["buttonColor"] = new JSONClass {
                    ["r"] = _buttonColor.r.ToString(),
                    ["g"] = _buttonColor.g.ToString(),
                    ["b"] = _buttonColor.b.ToString(),
                },
                ["buttonAlpha"] = _buttonAlpha.ToString(),
                ["textColor"] = new JSONClass {
                    ["r"] = _textColor.r.ToString(),
                    ["g"] = _textColor.g.ToString(),
                    ["b"] = _textColor.b.ToString(),
                },
                ["textAlpha"] = _textAlpha.ToString(),
                ["hidden"] = _hidden.ToString()
            };

            return jsonClass;
        }

        public void RestoreFromJSON(JSONClass buttonJson, string subScenePrefix, bool isMerge, bool setMissingToDefault) {
            buttonTrigger.RestoreFromJSON(buttonJson, subScenePrefix, isMerge, setMissingToDefault);
            buttonAltTrigger.RestoreFromJSON(buttonJson, subScenePrefix, isMerge, setMissingToDefault);
            _buttonLabel = buttonJson["buttonLabel"];
            _buttonAltLabel = buttonJson["buttonAltLabel"];
            JSONNode jsonButtonColor = buttonJson["buttonColor"];
            _buttonColor = ParseRGBJSONFloats(jsonButtonColor);
            _buttonColor.a = _buttonAlpha = float.Parse(buttonJson["buttonAlpha"]);
            JSONNode jsonTextColor = buttonJson["textColor"];
            _textColor = ParseRGBJSONFloats(jsonTextColor);
            _textColor.a = _textAlpha = float.Parse(buttonJson["textAlpha"]);
            _hidden = bool.Parse(buttonJson["hidden"]);
            OnUpdate();
        }

        private Color ParseRGBJSONFloats(JSONNode jsonNode) {
            return new Color(float.Parse(jsonNode["r"]), float.Parse(jsonNode["g"]), float.Parse(jsonNode["b"]));
        }
    }
}