﻿using AssetBundles;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;

namespace AwasUI {
    internal class AwasUI : MVRScript {
        private const float LOOK_AWAY_THRESHOLD = 30f;
        private const float LOOK_UP_DOWN_THRESHOLD = 50f;
        private const int MINI_CAMERA_CULLED_LAYER = 8;
        private const int MINI_CAMERA_NOT_CULLED_LAYER = 1;

        // IMPORTANT - DO NOT make custom enums. The dynamic C# complier crashes Unity when it encounters these for
        // some reason

        private Canvas worldSpaceUiCanvas;
        private Canvas worldSpaceVrHintCanvas;
        private Canvas camAnchorCanvas;
        private UIDynamicButton camIsRenderingToDesktopIndicator;

        private bool isVrActive = XRSettings.enabled;
        bool isWindowCamRenderers = true;

        private AwasUIButton[] buttons;
        private AwasUIButton currentConfigButton;
        private AwasUIButton buttonDownInstance;

        private UIDynamicButton triggerButton;
        private UIDynamicButton triggerAltButton;
        private InputField buttonLabelInput;
        private InputField buttonAltLabelInput;
        private UIDynamicColorPicker buttonColorPicker;
        private UIDynamicSlider buttonAlphaSlider;
        private UIDynamicColorPicker textColorPicker;
        private UIDynamicSlider textAlphaSlider;
        private UIDynamicToggle buttonHiddenToggle;
        private UIDynamicButton hintButton;
        private UIDynamicToggle camToDesktopToggle;
        private UIDynamicToggle uiIsAlwaysOnTopToggle;
        private UIDynamicToggle uiIsVisibleInMiniCameraToggle;
        private UIDynamicButton displayRotateButton;

        private float hintButtonColorSummand = 1;
        private Vector2[] buttonPositions;

        private EventTrigger camPrevTrigger;
        private EventTrigger camNextTrigger;

        private Atom windowCamera = null;
        private FreeControllerV3 displayControl = null;
        private FreeControllerV3 windowCameraControl = null;

        private JSONStorableBool cameraOnStorable = null;
        private JSONStorableBool cameraWillRenderToDesktop = null;
        private JSONStorableBool uiIsAlwaysOnTop = null;
        private JSONStorableBool uiIsVisibleInMiniCameraDisplay = null;
        private JSONStorableString miniCameraRenderResolution = null;
        private JSONStorableString miniCameraDisplayRotation;
        private bool initCameraOnValue = false;
        private Renderer[] windowCameraRenderers;
        private Transform[] renderersOriginalParents;
        private Shader unlitOverlayShader;
        private Shader unlitTextureShader;
        private bool isMiniCameraDisplayRotated = false;
        private float miniCameraDisplayRotationXRotation = 0;
        private float miniCameraDisplayRotationZRotation = 90;
        private bool lowerFrequencyChecks = true;
        private bool cameraLookingAtUi = true;

        private const string CAM_DISPLAY_ROTATION_AUTO = "Auto";
        private const string CAM_DISPLAY_ROTATION_LANDSCAPE = "Landscape";
        private const string CAM_DISPLAY_ROTATION_PORTRAIT = "Portrait";
        List<string> miniCameraDisplayRotationValues = new List<string> { 
            CAM_DISPLAY_ROTATION_AUTO,
            CAM_DISPLAY_ROTATION_LANDSCAPE,
            CAM_DISPLAY_ROTATION_PORTRAIT 
        };

        private Camera mainWindowCamera;

        private float mainButtonWidth;
        private float mainButtonHeight;
        private float camButtonHeight;
        private float camButtonWidth;
        private int mainButtonFontSize;
        private int camButtonFontSize;

        private float mainColumn1Pos;
        private float mainColumn2Pos;
        private float camColumn3Pos;
        private float camColumn4Pos;

        private float mainRow1Pos;
        private float mainRow2Pos;
        private float mainRow3Pos;

        private float camRow1Pos;
        private float camRow2Pos;
        private float camRow3Pos;
        private float camRow4Pos;

        private Color mainButtonColor;
        private Color mainButtonTextColor;
        private Color camButtonColor;
        private Color camButtonTextColor;

        #region Setup
        // IMPORTANT - DO NOT OVERRIDE Awake() as it is used internally by MVRScript - instead use Init() function which
        // is called right after creation
        public override void Init() {
            try {
                // put init code in here
                SimpleTriggerHandler.LoadAssets();

                if (XRSettings.enabled) {
                    isVrActive = true;
                }

                camPrevTrigger = new EventTrigger(this, "camPrevTrigger");
                camNextTrigger = new EventTrigger(this, "camNextTrigger");

                SetButtonLayoutAndAppearance();
                CreateDefaultButtons();
                CreateWorldSpaceUi();
                CreateConfigUi();
                CreatePluginAPI();
                SetupWindowCameraAndDisplayControl();
                ApplyStoredConfig();

                StartCoroutine(DeferredLoad());

                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            } catch (Exception e) {
                SuperController.LogError("Init exception: " + e);
            }
        }

        private IEnumerator DeferredLoad() {
            yield return new WaitForEndOfFrame();
            if (this == null) yield break;

            while (SuperController.singleton.isLoading) {
                yield return 0;
                if (this == null) yield break;
            }

            base.containingAtom.RestoreFromLast(this);
        }

        private void ApplyStoredConfig() {
            SetCameraTextureRenderResolution(int.Parse(miniCameraRenderResolution.val));
            SetMiniCameraRotationState(miniCameraDisplayRotation.val);
            SetUiForCurrentButton(buttons[0]);
            camToDesktopToggle.toggle.isOn = cameraWillRenderToDesktop.val;
            uiIsAlwaysOnTopToggle.toggle.isOn = uiIsAlwaysOnTop.val;
            uiIsVisibleInMiniCameraToggle.toggle.isOn = uiIsVisibleInMiniCameraDisplay.val;
        }

        private IEnumerator CheckEvery100millis() {
            while (lowerFrequencyChecks) {
                UpdateIsMiniCameraDisplayRotated(windowCameraControl);
                if (isVrActive) ShowVrHintWhenCameraNotLookingAtUi();
                if (!cameraLookingAtUi && buttonDownInstance != null) {
                    buttonDownInstance.ResetUiButtonToDefaultState();
                    buttonDownInstance = null;
                }
                yield return new WaitForSeconds(0.1f);
            }
        }

        private void SetupWindowCameraAndDisplayControl() {
            windowCamera = SuperController.singleton.GetAtomByUid("WindowCamera");
            windowCameraControl = windowCamera.freeControllers.First(fc => fc.name == "control");
            SuperController.singleton.StartCoroutine(CheckEvery100millis());

            for (int i = 0; i < windowCamera.freeControllers.Length; i++) {
                if (windowCamera.freeControllers[i].name == "DisplayControl") {
                    displayControl = windowCamera.freeControllers[i];
                }
            }

            JSONStorable cameraControlStorable = windowCamera.GetStorableByID("CameraControl");
            cameraOnStorable = (JSONStorableBool)cameraControlStorable.GetParam("cameraOn");
            initCameraOnValue = cameraOnStorable.val;
            cameraOnStorable.val = true;

            JSONStorable cameraScaleStorable = windowCamera.GetStorableByID("scale");
            JSONStorableFloat scale = (JSONStorableFloat)cameraScaleStorable.GetParam("scale");
            scale.val = 1.0f;

            windowCameraRenderers = windowCamera.GetComponentsInChildren<Renderer>();
            renderersOriginalParents = new Transform[windowCameraRenderers.Length];
            ApplyToMiniCamRenderer(renderer => {
                renderersOriginalParents[0] = renderer.transform.parent;
                renderer.transform.SetParent(camAnchorCanvas.transform);
                renderer.gameObject.layer = MINI_CAMERA_CULLED_LAYER;
            });
            windowCameraRenderers[1].gameObject.SetActive(false);
            renderersOriginalParents[1] = windowCameraRenderers[1].transform.parent;

            try {
                Shader[] shaders = Resources.FindObjectsOfTypeAll<Shader>();
                unlitOverlayShader = Array.Find(shaders, s => s.name == "Unlit/UnlitOverlayShader");
                unlitTextureShader = Array.Find(shaders, s => s.name == "Unlit/Texture");
                windowCameraRenderers[0].material.shader = unlitOverlayShader;
            } catch (Exception e) {
                SuperController.LogMessage(e.Message);
            }

            foreach (Camera sceneCamera in Camera.allCameras) {
                if (sceneCamera.name == "MainWindowCamera") {
                    mainWindowCamera = sceneCamera;
                }
            }
            mainWindowCamera.enabled = false;
        }

        private void SetButtonLayoutAndAppearance() {
            float mainButtonPadding = isVrActive ? 2f : 5f;
            float camButtonPadding = isVrActive ? 1.5f : 4f;

            mainButtonWidth = isVrActive ? 45f : 90f;
            mainButtonHeight = isVrActive ? 35f : 40f;
            mainButtonFontSize = isVrActive ? 10 : 18;
            camButtonFontSize = isVrActive ? 10 : 10;
            camButtonHeight = isVrActive ? 23f : 25.3f;
            camButtonWidth = isVrActive ? 30 : 40;

            mainColumn1Pos = 0;
            mainColumn2Pos = mainColumn1Pos + mainButtonWidth + mainButtonPadding;
            camColumn3Pos = mainColumn2Pos + mainButtonWidth + mainButtonPadding;
            camColumn4Pos = camColumn3Pos + camButtonWidth + camButtonPadding;

            mainRow1Pos = 15;
            mainRow2Pos = mainRow1Pos - mainButtonHeight - mainButtonPadding;
            mainRow3Pos = mainRow2Pos - mainButtonHeight - mainButtonPadding;

            camRow1Pos = mainRow1Pos;
            camRow2Pos = camRow1Pos - camButtonHeight - camButtonPadding;
            camRow3Pos = camRow2Pos - camButtonHeight - camButtonPadding;
            camRow4Pos = mainRow3Pos;

            buttonPositions = new Vector2[6];
            buttonPositions[0] = new Vector2(mainColumn1Pos, mainRow1Pos);
            buttonPositions[1] = new Vector2(mainColumn2Pos, mainRow1Pos);
            buttonPositions[2] = new Vector2(mainColumn1Pos, mainRow2Pos);
            buttonPositions[3] = new Vector2(mainColumn2Pos, mainRow2Pos);
            buttonPositions[4] = new Vector2(mainColumn1Pos, mainRow3Pos);
            buttonPositions[5] = new Vector2(mainColumn2Pos, mainRow3Pos);
            
            mainButtonColor = new Color(0, 0, 0, 0.25f);
            mainButtonTextColor = new Color(1, 0.75f, 0.95f, 0.25f);
            camButtonColor = new Color(0, 0, 0, 0.25f);
            camButtonTextColor = new Color(1, 1, 1, 1);
        }
        #endregion

        #region API
        private void CreatePluginAPI() {
            RegisterBool(cameraWillRenderToDesktop);

            foreach (AwasUIButton button in buttons) {
                JSONStorableString storableLabel = new JSONStorableString(button.configUiName + " button label", "", (newLabel) => {
                    button.buttonLabel = newLabel;
                }) { isRestorable = false, isStorable = false };
                RegisterString(storableLabel);

                JSONStorableColor storableButtonColor = new JSONStorableColor(button.configUiName + " button color", ColorToHSV(button.buttonColor), (h, s, v) => {
                    button.buttonColor = HSVColorPicker.HSVToRGB(new HSVColor() { H = h, S = s, V = v});
                }) { isRestorable = false, isStorable = false };
                RegisterColor(storableButtonColor);

                JSONStorableFloat storableButtonAlpha = new JSONStorableFloat(button.configUiName + " button alpha", 0.75f, (newAlpha) => {
                    button.buttonAlpha = newAlpha;
                }, 0, 1) { isRestorable = false, isStorable = false };
                RegisterFloat(storableButtonAlpha);

                JSONStorableColor storableTextColor = new JSONStorableColor(button.configUiName + " text color", ColorToHSV(button.textColor), (h, s, v) => {
                    button.textColor = HSVColorPicker.HSVToRGB(new HSVColor() { H = h, S = s, V = v });
                }) { isRestorable = false, isStorable = false };
                RegisterColor(storableTextColor);

                JSONStorableFloat storableTextAlpha = new JSONStorableFloat(button.configUiName + " text alpha", 0.75f, (newAlpha) => {
                    button.textAlpha = newAlpha;
                }, 0, 1) { isRestorable = false, isStorable = false };
                RegisterFloat(storableTextAlpha);

                JSONStorableBool storableIsHidden = new JSONStorableBool(button.configUiName + " hidden", false, (newIsHidden) => {
                    button.hidden = newIsHidden;
                }) { isRestorable = false, isStorable = false };
                RegisterBool(storableIsHidden);
            }
        }
        #endregion

        private void CreateDefaultButtons() {
            AwasUIButton topLeftButton = new AwasUIButton(this, "Top left", "TL", "TL alt", buttonPositions[0].x, buttonPositions[0].y);
            AwasUIButton topRightButton = new AwasUIButton(this, "Top right", "TR", "TR alt", buttonPositions[1].x, buttonPositions[1].y);
            AwasUIButton bottomLeftButton = new AwasUIButton(this, "Bottom left", "BL", "BL alt", buttonPositions[2].x, buttonPositions[2].y);
            AwasUIButton bottomRightButton = new AwasUIButton(this, "Bottom right", "BR", "BR alt", buttonPositions[3].x, buttonPositions[3].y);
            AwasUIButton bottomLeftUtilButton = new AwasUIButton(this, "Bottom left util", "BL", "BL alt", buttonPositions[4].x, buttonPositions[4].y, mainButtonWidth, camButtonHeight);
            AwasUIButton bottomRightUtilButton = new AwasUIButton(this, "Bottom right util", "BR", "BR alt", buttonPositions[5].x, buttonPositions[5].y, mainButtonWidth, camButtonHeight);

            buttons = new AwasUIButton[] { topLeftButton, topRightButton, bottomLeftButton, bottomRightButton, bottomLeftUtilButton, bottomRightUtilButton };
        }

        #region ConfigUi
        private void SetMiniCameraRotationState(string rotation) {
            displayRotateButton.label = rotation.Remove(1);
            if (rotation == CAM_DISPLAY_ROTATION_LANDSCAPE) {
                isMiniCameraDisplayRotated = false;
                miniCameraDisplayRotationXRotation = 90;
                miniCameraDisplayRotationZRotation = 0;
            } else if (rotation == CAM_DISPLAY_ROTATION_PORTRAIT) {
                isMiniCameraDisplayRotated = true;
                miniCameraDisplayRotationXRotation = 180;
                miniCameraDisplayRotationZRotation = 90;
            } else if (rotation == CAM_DISPLAY_ROTATION_AUTO) {
                UpdateIsMiniCameraDisplayRotated(windowCameraControl);
            }
        }

        private void SetCameraWillRenderToDesktop(bool willRender) {
            mainWindowCamera.enabled = willRender;
            camIsRenderingToDesktopIndicator.gameObject.SetActive(willRender);
            cameraWillRenderToDesktop.val = willRender;
        }

        private void SetCameraTextureRenderResolution(int textureHeight) {
            const float ASPECT_RATIO = 1.77f;
            int textureWidth = (int)(textureHeight * ASPECT_RATIO);

            RenderTexture windowCameraRenderTexture = (RenderTexture)windowCameraRenderers[0].material.mainTexture;
            windowCameraRenderTexture.Release();
            windowCameraRenderTexture.width = textureWidth;
            windowCameraRenderTexture.height = textureHeight;
            windowCameraRenderTexture.Create();
        }

        private void CreateConfigUi() {
            cameraWillRenderToDesktop = new JSONStorableBool("cameraWillRenderToDesktop", false) { isRestorable = true, isStorable = true };
            miniCameraRenderResolution = new JSONStorableString("miniCameraRenderResolution", "512") { isRestorable = true, isStorable = true };
            miniCameraDisplayRotation = new JSONStorableString("miniCameraDisplayRotation", CAM_DISPLAY_ROTATION_AUTO) { isRestorable = true, isStorable = true };
            uiIsVisibleInMiniCameraDisplay = new JSONStorableBool("uiIsVisibleInMiniCameraDisplay", false) { isRestorable = true, isStorable = true };
            uiIsAlwaysOnTop = new JSONStorableBool("uiIsAlwaysOnTop", true) { isRestorable = true, isStorable = true };

            // Left display settings
            CreateConfigUiLabel("Wrist display settings");
            CreateConfigUiSpcer(10f);

            List<string> resolutionChoices = new List<string>{ "256", "512", "768", "1024" };
            JSONStorableStringChooser resolutionSelector = new JSONStorableStringChooser("Resolution", resolutionChoices, miniCameraRenderResolution.val, "Resolution") {
                isRestorable = false, isStorable = false
            };
            resolutionSelector.setCallbackFunction += (string choice) => {
                SetCameraTextureRenderResolution(int.Parse(choice));
                miniCameraRenderResolution.val = choice;
            };
            UIDynamicPopup resolutionSelectorPopup = CreateScrollablePopup(resolutionSelector, false);
            resolutionSelectorPopup.labelTextColor = Color.black;
            
            miniCameraRenderResolution.setCallbackFunction += (string newValue) => {
                resolutionSelector.val = newValue;
            };

            JSONStorableStringChooser displayRotationSelector = new JSONStorableStringChooser("Display rotation", miniCameraDisplayRotationValues, miniCameraDisplayRotation.val, "Display rotation") {
                isRestorable = false, isStorable = false
            };
            RegisterStringChooser(displayRotationSelector);
            displayRotationSelector.setCallbackFunction += (string choice) => {
                miniCameraDisplayRotation.val = choice;
                SetMiniCameraRotationState(choice);
            };
            UIDynamicPopup displayRotationSelectorPopup = CreateScrollablePopup(displayRotationSelector, false);
            displayRotationSelectorPopup.labelTextColor = Color.black;

            miniCameraDisplayRotation.setCallbackFunction += (string newValue) => {
                displayRotationSelector.val = newValue;
            };

            // Left camera settings
            CreateConfigUiSpcer(10f); 
            CreateConfigUiLabel("Camera settings");
            CreateConfigUiSpcer(10f);

            camToDesktopToggle = CreateConfigUiToggle("Camera renders to desktop window");
            camToDesktopToggle.toggle.onValueChanged.AddListener(willRender => SetCameraWillRenderToDesktop(willRender));

            UIDynamicButton camPrevButton = CreateButton("Cam prev trigger");
            UIDynamicButton camNextButton = CreateButton("Cam next trigger");
            camPrevButton.button.onClick.AddListener(camPrevTrigger.OpenPanelActionStart);
            camNextButton.button.onClick.AddListener(camNextTrigger.OpenPanelActionStart);

            CreateConfigUiSpcer(10f);
            CreateConfigUiLabel("UI display settings"); 
            CreateConfigUiSpcer(10f);

            uiIsAlwaysOnTopToggle = CreateConfigUiToggle("Mini display is always on top");
            uiIsAlwaysOnTopToggle.toggle.onValueChanged.AddListener(uiIsAlwaysOnTop => ToggleMiniCameraDisplayFrontBack(uiIsAlwaysOnTop));
            
            uiIsVisibleInMiniCameraToggle = CreateConfigUiToggle("UI is visible in mini camera display");
            uiIsVisibleInMiniCameraToggle.toggle.onValueChanged.AddListener(uiIsVisible => ToggleUiVisibleInMiniCameraDisplay(uiIsVisible));

            // Right button settings
            CreateConfigUiLabel("Button settings", true);
            CreateConfigUiSpcer(10f, true);

            List<string> buttonChoices = new List<string>();
            foreach (AwasUIButton button in buttons) {
                buttonChoices.Add(button.configUiName);
            }

            JSONStorableStringChooser buttonSelector = new JSONStorableStringChooser("Buttons", buttonChoices, buttons[0].configUiName, "Button") {
                isRestorable = false, isStorable = false
            };
            buttonSelector.setCallbackFunction += (string choice) => SetUiForCurrentButton(GetButtonFromName(choice));
            UIDynamicPopup buttonSelectorPopup = CreateScrollablePopup(buttonSelector, true);
            buttonSelectorPopup.labelTextColor = Color.black;

            buttonHiddenToggle = CreateConfigUiToggle("Button hidden", true);
            triggerButton = CreateConfigUiTriggersButton("Triggers", true);
            triggerAltButton = CreateConfigUiTriggersButton("Long press triggers", true);
            CreateConfigUiLabel("Button label", true);
            buttonLabelInput = CreateConfigUiTextInput(true);
            CreateConfigUiLabel("Button long press label", true);
            buttonAltLabelInput = CreateConfigUiTextInput(true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Button color", true);
            buttonColorPicker = CreateConfigUiColorChooser(ColorToHSV(mainButtonColor), true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Button alpha", true);
            buttonAlphaSlider = CreateConfigUiAlphaSlider(true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Text color", true);
            textColorPicker = CreateConfigUiColorChooser(ColorToHSV(mainButtonTextColor), true);
            CreateConfigUiSpcer(10f, true);
            CreateConfigUiLabel("Text alpha", true);
            textAlphaSlider = CreateConfigUiAlphaSlider(true);

            SetUiForCurrentButton(buttons[0]);
        }

        private void SetUiForCurrentButton(AwasUIButton buttonToConfig) {
            currentConfigButton = buttonToConfig;

            triggerButton.button.onClick.RemoveAllListeners();
            triggerButton.button.onClick.AddListener(currentConfigButton.buttonTrigger.OpenPanelActionStart);

            triggerAltButton.button.onClick.RemoveAllListeners();
            triggerAltButton.button.onClick.AddListener(currentConfigButton.buttonAltTrigger.OpenPanelActionStart);

            buttonLabelInput.onValueChanged.RemoveAllListeners();
            buttonLabelInput.onValueChanged.AddListener(newLabel => currentConfigButton.buttonLabel = newLabel);
            buttonLabelInput.text = currentConfigButton.buttonLabel;

            buttonAltLabelInput.onValueChanged.RemoveAllListeners();
            buttonAltLabelInput.onValueChanged.AddListener(newLabel => currentConfigButton.buttonAltLabel = newLabel);
            buttonAltLabelInput.text = currentConfigButton.buttonAltLabel;

            buttonColorPicker.colorPicker.onColorChangedHandlers = null;
            buttonColorPicker.colorPicker.onColorChangedHandlers += (color) => currentConfigButton.buttonColor = color;
            buttonColorPicker.colorPicker.SetHSV(ColorToHSV(currentConfigButton.buttonColor));

            buttonAlphaSlider.slider.onValueChanged.RemoveAllListeners();
            buttonAlphaSlider.slider.onValueChanged.AddListener(newAlpha => currentConfigButton.buttonAlpha = newAlpha);
            buttonAlphaSlider.slider.value = currentConfigButton.buttonAlpha;

            textColorPicker.colorPicker.onColorChangedHandlers = null;
            textColorPicker.colorPicker.onColorChangedHandlers += (color) => currentConfigButton.textColor = color;
            textColorPicker.colorPicker.SetHSV(ColorToHSV(currentConfigButton.textColor));

            textAlphaSlider.slider.onValueChanged.RemoveAllListeners();
            textAlphaSlider.slider.onValueChanged.AddListener(newAlpha => currentConfigButton.textAlpha = newAlpha);
            textAlphaSlider.slider.value = currentConfigButton.textAlpha;

            buttonHiddenToggle.toggle.onValueChanged.RemoveAllListeners();
            buttonHiddenToggle.toggle.onValueChanged.AddListener(isHidden => currentConfigButton.hidden = isHidden);
            buttonHiddenToggle.toggle.isOn = currentConfigButton.hidden;
        }

        private AwasUIButton GetButtonFromName(string name) {
            AwasUIButton button = buttons[0];
            foreach (AwasUIButton awasUIButton in buttons) {
                if (awasUIButton.configUiName == name) {
                    button = awasUIButton;
                    break;
                }
            }
            return button;
        }

        private UIDynamicToggle CreateConfigUiToggle(string toggleLabel, bool rightSide = false) {
            return CreateToggle(new JSONStorableBool(toggleLabel, false) { isRestorable = false, isStorable = false }, rightSide);
        }

        private UIDynamicSlider CreateConfigUiAlphaSlider(bool rightSide = false) {
            UIDynamicSlider slider = CreateSlider(new JSONStorableFloat("buttonAlpha", 0.5f, 0, 1) { isStorable = false, isRestorable = false }, rightSide);
            slider.labelText.text = "Alpha";
            return slider;
        }

        private UIDynamicColorPicker CreateConfigUiColorChooser(HSVColor defaultColor, bool rightSide = false) {
            UIDynamicColorPicker colorPicker = CreateColorPicker(new JSONStorableColor("buttonColor", defaultColor) { isRestorable = false, isStorable = false }, rightSide);
            colorPicker.showLabel = false;
            return colorPicker;
        }

        private InputField CreateConfigUiTextInput(bool rightSide = false) {
            JSONStorableString storableString = new JSONStorableString("", "");
            UIDynamicTextField inputTextContainer = CreateTextField(storableString, rightSide);
            inputTextContainer.backgroundColor = Color.white;
            inputTextContainer.UItext.GetComponent<RectTransform>().anchoredPosition = new Vector2(10f, -10f);
            ScrollRect[] scrollRects = inputTextContainer.UItext.transform.GetComponentsInParent<ScrollRect>(true);
            for (int i = 0; i < scrollRects.Length; i++) {
                scrollRects[i].horizontal = false;
                scrollRects[i].vertical = false;
            }

            LayoutElement labelLayout = inputTextContainer.GetComponent<LayoutElement>();
            labelLayout.minHeight = labelLayout.preferredHeight = inputTextContainer.height = 50f;

            InputField input = inputTextContainer.gameObject.AddComponent<InputField>();
            input.textComponent = inputTextContainer.UItext;

            return input;
        }

        private UIDynamicButton CreateConfigUiTriggersButton(string buttonLabel, bool rightSide = false) {
            return CreateButton(buttonLabel, rightSide);
        }

        private void CreateConfigUiLabel(string uiLabel, bool rightSide = false) {
            JSONStorableString tempStorable = new JSONStorableString("", uiLabel);
            tempStorable.isRestorable = false;
            tempStorable.isStorable = false;

            UIDynamicTextField labelTextField = CreateTextField(tempStorable, rightSide);
            labelTextField.backgroundColor = new Color(1,1,1,0.25f);
            labelTextField.UItext.GetComponent<RectTransform>().anchoredPosition = new Vector2(10f, -10f);
            ScrollRect[] scrollRects = labelTextField.UItext.transform.GetComponentsInParent<ScrollRect>(true);
            for (int i = 0; i < scrollRects.Length; i++) {
                scrollRects[i].horizontal = false;
                scrollRects[i].vertical = false;
            }

            LayoutElement labelLayout = labelTextField.GetComponent<LayoutElement>();
            labelLayout.minHeight = labelLayout.preferredHeight = labelTextField.height = 50f;
        }

        private void CreateConfigUiSpcer(float height, bool rightSide = false) {
            UIDynamic spacer = CreateSpacer(rightSide);
            spacer.height = height;
        }

        private HSVColor ColorToHSV(Color color) {
            return HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
        }
        #endregion

        #region WorldSpaceUi
        private void CreateWorldSpaceUi() {
            CreateWorldSpaceCanvases();
            foreach (AwasUIButton button in buttons) {
                UIDynamicButton worldSpaceButton = CreateWorldSpaceButtonFromAwasButton(button);
                button.SetUiButton(worldSpaceButton);
                button.OnButtonDown(() => buttonDownInstance = button);
            }

            if (isVrActive) {
                hintButton = CreateWorldSpaceButton(worldSpaceVrHintCanvas.transform, 50f, 25f, 25f, mainRow2Pos);
                hintButton.label = "";
                hintButton.buttonColor = mainButtonColor;
                hintButton.buttonText.color = mainButtonTextColor;
                hintButton.buttonText.fontSize = 12;
                SuperController.singleton.StartCoroutine(VrHintButtonBreathingUpdate());
            }

            UIDynamicButton camPrevButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow1Pos, 12);
            UIDynamicButton camToggleButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow2Pos, camButtonFontSize);
            UIDynamicButton camNextButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow3Pos, 12);
            camIsRenderingToDesktopIndicator = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth / 3, camButtonHeight, camColumn4Pos, camRow2Pos);
            displayRotateButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, camButtonWidth, camButtonHeight, camColumn3Pos, camRow4Pos, 12);

            int rotationIndex = 0;
            displayRotateButton.button.onClick.AddListener(() => {
                rotationIndex = miniCameraDisplayRotationValues.IndexOf(miniCameraDisplayRotation.val) + 1;
                rotationIndex = rotationIndex > 2 ? 0 : rotationIndex;
                string rotation = miniCameraDisplayRotationValues[rotationIndex];
                miniCameraDisplayRotation.val = rotation;
                SetMiniCameraRotationState(rotation);
            });

            camPrevButton.label = "<";
            camNextButton.label = ">";
            camToggleButton.label = "Cam";
            camIsRenderingToDesktopIndicator.label = "";
            displayRotateButton.label = "";

            camToggleButton.button.onClick.AddListener(ToggleMiniCameraDisplay);
            camPrevButton.button.onClick.AddListener(() => {
                camPrevTrigger.Trigger();
                UpdateIsMiniCameraDisplayRotated(windowCameraControl);
            });
            camNextButton.button.onClick.AddListener(() => {
                camNextTrigger.Trigger();
                UpdateIsMiniCameraDisplayRotated(windowCameraControl);
            });

            displayRotateButton.buttonColor = camPrevButton.buttonColor = camToggleButton.buttonColor = camNextButton.buttonColor = camButtonColor;
            displayRotateButton.textColor = camPrevButton.textColor = camToggleButton.textColor = camNextButton.textColor = camButtonTextColor;
            
            camIsRenderingToDesktopIndicator.buttonColor = Color.green;
            camIsRenderingToDesktopIndicator.gameObject.SetActive(false);
        }

        private void ToggleMiniCameraDisplay() {
            cameraOnStorable.val = !cameraOnStorable.val;
            ApplyToMiniCamRenderer((Renderer renderer) => renderer.gameObject.SetActive(!renderer.gameObject.activeSelf));
        }

        private void ToggleMiniCameraDisplayFrontBack(bool isUiOnTop) {
            uiIsAlwaysOnTop.val = isUiOnTop;
            ApplyToMiniCamRenderer((Renderer renderer) => renderer.material.shader = isUiOnTop ? unlitOverlayShader : unlitTextureShader);
        }

        private void ToggleUiVisibleInMiniCameraDisplay(bool isUiVisible) {
            uiIsVisibleInMiniCameraDisplay.val = isUiVisible;
            if (!isUiVisible) {
                SetWorldSpaceUiObjectsLayer(MINI_CAMERA_CULLED_LAYER);
            } else {
                SetWorldSpaceUiObjectsLayer(MINI_CAMERA_NOT_CULLED_LAYER);
            }
        }

        private void SetWorldSpaceUiObjectsLayer(int layer) {
            if (worldSpaceVrHintCanvas != null) worldSpaceVrHintCanvas.gameObject.layer = layer;
            worldSpaceUiCanvas.gameObject.layer = layer;
            camAnchorCanvas.gameObject.layer = layer;
            ApplyToMiniCamRenderer((Renderer renderer) => renderer.gameObject.layer = layer);
        }

        private void ApplyToMiniCamRenderer(Action<Renderer> applyThisToRenderer) {
            applyThisToRenderer(windowCameraRenderers[0]);
        }

        private void CreateWorldSpaceCanvases() {
            worldSpaceUiCanvas = CreateCanvas(true);
            camAnchorCanvas = CreateCanvas(false);

            if (isVrActive) {
                worldSpaceVrHintCanvas = CreateCanvas(false);
            }
        }

        private Canvas CreateCanvas(bool withRaycaster) {
            GameObject newGameObject = new GameObject() { layer = MINI_CAMERA_CULLED_LAYER };

            float localScale = 0.001f * SuperController.singleton.worldScale;
            Canvas newCanvas = newGameObject.AddComponent<Canvas>();
            newCanvas.renderMode = RenderMode.WorldSpace;
            newCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            SuperController.singleton.AddCanvas(newCanvas);

            CanvasScaler uiCanvasScaler = newGameObject.AddComponent<CanvasScaler>();
            if (isVrActive) {
                uiCanvasScaler.dynamicPixelsPerUnit = 4f;
            } else {
                uiCanvasScaler.dynamicPixelsPerUnit = 1f;
            }

            if (withRaycaster) {
                newGameObject.AddComponent<GraphicRaycaster>();
            }

            return newCanvas;
        }

        private UIDynamicButton CreateWorldSpaceButtonFromAwasButton(AwasUIButton button) {
            float buttonWidth = mainButtonWidth;
            float buttonHeight = mainButtonHeight;
            if (button.buttonWidth > 0 && button.buttonHeight > 0) {
                buttonWidth = button.buttonWidth;
                buttonHeight = button.buttonHeight;
            }
            UIDynamicButton uiButton = CreateWorldSpaceButton(worldSpaceUiCanvas.transform, buttonWidth, buttonHeight, button.worldXpos, button.worldYpos, mainButtonFontSize);
            uiButton.buttonColor = button.buttonColor;
            uiButton.label = button.buttonLabel;
            uiButton.buttonText.color = button.textColor;
            return uiButton;
        }

        private UIDynamicButton CreateWorldSpaceButton(Transform parent, float width, float height, float xPos, float yPos, int fontSize = 18) {
            Transform transform = Instantiate(manager.configurableButtonPrefab.transform);
            transform.SetParent(parent, false);
            transform.gameObject.SetActive(true);
            transform.localPosition = new Vector3(xPos, yPos, 0f);

            UIDynamicButton uiButton = transform.GetComponent<UIDynamicButton>();
            uiButton.buttonText.fontSize = fontSize;

            uiButton.button.GetComponent<RectTransform>().pivot = new Vector2(0, 1);
            uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            return uiButton;
        }
        #endregion

        #region Update
        private void UpdateCanvasLocalScale() {
            float localScale = 0.001f * SuperController.singleton.worldScale;
            worldSpaceUiCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            worldSpaceVrHintCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
            camAnchorCanvas.transform.localScale = new Vector3(localScale, localScale, localScale);
        }

        private void UpdateDesktopUiPosition() {
            Transform lookCameraTransform = SuperController.singleton.lookCamera.transform;
            Vector3 lookAtPosition = lookCameraTransform.position + lookCameraTransform.forward * 100f;
            worldSpaceUiCanvas.transform.LookAt(lookAtPosition, lookCameraTransform.up);
            camAnchorCanvas.transform.position = worldSpaceUiCanvas.transform.position = Camera.main.ViewportToWorldPoint(new Vector3(0.875f, 0.2f, 1.7f));
            camAnchorCanvas.transform.rotation = worldSpaceUiCanvas.transform.rotation;

            ApplyToMiniCamRenderer(renderer => {
                if (isMiniCameraDisplayRotated) {
                    renderer.transform.position = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.16f) + (camAnchorCanvas.transform.up * -0.04f);
                    renderer.transform.rotation = camAnchorCanvas.transform.rotation * Quaternion.Euler(miniCameraDisplayRotationXRotation, 270, miniCameraDisplayRotationZRotation);
                } else {
                    renderer.transform.position = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.22f) + (camAnchorCanvas.transform.up * -0.04f);
                    renderer.transform.rotation = worldSpaceUiCanvas.transform.rotation * Quaternion.Euler(90, 180, 0);
                }
            });
    }

        private bool isWorldSpaceUiLerping = false;
        private short prevCamRotation;
        private short currCamRotation;
        private Quaternion prevWorldSpaceUiRotation;
        private Quaternion prevWindowRendererRotation;
        private float lerpCount = 0;
        private void UpdateVrUiPositionAndVisibility() {
            prevCamRotation = currCamRotation;
            Transform worldSpaceTransform = worldSpaceUiCanvas.transform;
            Transform lookCameraTransform = SuperController.singleton.lookCamera.transform;
            Transform leftHandTransform = SuperController.singleton.leftHand.transform;
            float worldScale = SuperController.singleton.worldScale;

            Vector3 offsetPosition = (worldSpaceTransform.up * 0.15f * worldScale) + (worldSpaceTransform.right * -0.05f * worldScale) + (worldSpaceTransform.forward * -0.03f * worldScale);
            worldSpaceTransform.position = SuperController.singleton.leftHand.position + offsetPosition;

            float lookCameraUpDownAngle = Vector3.Angle(SuperController.singleton.transform.up, lookCameraTransform.up);
            float handToCameraDist2D = Vector3.Distance(new Vector3(leftHandTransform.position.x, 0, leftHandTransform.position.z), new Vector3(lookCameraTransform.position.x, 0, lookCameraTransform.position.z)) * worldScale;

            if (!isWorldSpaceUiLerping) {
                prevWorldSpaceUiRotation = worldSpaceTransform.rotation;
                prevWindowRendererRotation = windowCameraRenderers[0].transform.rotation;
            }

            bool uiIsUsingCameraRotation;
            Quaternion newWorldSpaceUiRotation;
            if (lookCameraUpDownAngle > LOOK_UP_DOWN_THRESHOLD || handToCameraDist2D <= 0.3f) {
                newWorldSpaceUiRotation = lookCameraTransform.rotation;
                uiIsUsingCameraRotation = true;
                currCamRotation = 0;
            } else {
                currCamRotation = 1;
                worldSpaceTransform.LookAt(lookCameraTransform.position);
                worldSpaceTransform.rotation *= Quaternion.Euler(0, 180, 0);
                newWorldSpaceUiRotation = worldSpaceTransform.rotation;
                worldSpaceTransform.rotation = prevWorldSpaceUiRotation;
                uiIsUsingCameraRotation = false;
            }

            if (currCamRotation != prevCamRotation) {
                lerpCount = 0;
                isWorldSpaceUiLerping = true;
            }

            TransitionToNewRotation(worldSpaceTransform, prevWorldSpaceUiRotation, newWorldSpaceUiRotation);

            camAnchorCanvas.transform.position = worldSpaceVrHintCanvas.transform.position = worldSpaceTransform.position;
            camAnchorCanvas.transform.rotation = worldSpaceVrHintCanvas.transform.rotation = worldSpaceTransform.rotation;

            Quaternion newWindowRendererRotation;
            Vector3 camDisplayForwardOffset;
            Vector3 camDisplayPosition;

            if (!isMiniCameraDisplayRotated) {
                camDisplayForwardOffset = uiIsUsingCameraRotation ? Vector3.zero : (camAnchorCanvas.transform.forward * -0.05f * worldScale);
                camDisplayPosition = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.22f * worldScale) + (camAnchorCanvas.transform.up * -0.04f * worldScale) + camDisplayForwardOffset;
                float camDisplayYRotation = uiIsUsingCameraRotation ? 180 : 150;
                newWindowRendererRotation = camAnchorCanvas.transform.rotation * Quaternion.Euler(miniCameraDisplayRotationXRotation, camDisplayYRotation, miniCameraDisplayRotationZRotation);
            } else {
                camDisplayForwardOffset = camAnchorCanvas.transform.forward * 0.0f * worldScale;
                camDisplayPosition = camAnchorCanvas.transform.position + (camAnchorCanvas.transform.right * -0.15f * worldScale) + (camAnchorCanvas.transform.up * -0.04f * worldScale) + camDisplayForwardOffset;
                newWindowRendererRotation = camAnchorCanvas.transform.rotation * Quaternion.Euler(miniCameraDisplayRotationXRotation, 270, miniCameraDisplayRotationZRotation);
            }

            ApplyToMiniCamRenderer(renderer => {
                TransitionToNewRotation(renderer.transform, prevWindowRendererRotation, newWindowRendererRotation);
                renderer.transform.position = camDisplayPosition;
            });
        }

        private void ShowVrHintWhenCameraNotLookingAtUi() {
            Transform lookCameraTransform = SuperController.singleton.lookCamera.transform;
            Transform worldSpaceTransform = worldSpaceUiCanvas.transform;
            float worldScale = SuperController.singleton.worldScale;
            Transform leftHandTransform = SuperController.singleton.leftHand.transform;

            Vector3 lookCameraOffset = (lookCameraTransform.forward * worldScale) + (lookCameraTransform.right * -0.1f * worldScale) + (lookCameraTransform.up * 0.3f * worldScale);
            float lookCameraAngleToUiCanvas = Vector3.Angle(worldSpaceTransform.position - lookCameraTransform.position, lookCameraOffset);
            float handToLookCameraDistance = Vector3.Distance(leftHandTransform.position, lookCameraTransform.position) * worldScale;

            if (lookCameraAngleToUiCanvas < LOOK_AWAY_THRESHOLD && handToLookCameraDistance > 0.3f) {
                cameraLookingAtUi = true;
                worldSpaceUiCanvas.gameObject.SetActive(true);
                worldSpaceVrHintCanvas.gameObject.SetActive(false);
            } else {
                cameraLookingAtUi = false;
                worldSpaceUiCanvas.gameObject.SetActive(false);
                worldSpaceVrHintCanvas.gameObject.SetActive(true);
            }
        }

        private void UpdateIsMiniCameraDisplayRotated(FreeControllerV3 windowCameraController) {
            float ZRotation = 180 - windowCameraController.transform.rotation.eulerAngles.z;
            float ZRotationAbs = Math.Abs(ZRotation);
            if (miniCameraDisplayRotation.val == CAM_DISPLAY_ROTATION_AUTO) {
                if (ZRotationAbs > 55 && ZRotationAbs < 125) {
                    isMiniCameraDisplayRotated = true;
                    miniCameraDisplayRotationXRotation = ZRotation >= 0 ? 180 : 0;
                    miniCameraDisplayRotationZRotation = 90;
                } else {
                    isMiniCameraDisplayRotated = false;
                    miniCameraDisplayRotationXRotation = ZRotationAbs > 125 ? 90 : -90;
                    miniCameraDisplayRotationZRotation = miniCameraDisplayRotationXRotation == 90 ? 0 : 180;
                }
            }
        }

        private void TransitionToNewRotation(Transform transform, Quaternion oldRotation, Quaternion newRotation) {
            if (isWorldSpaceUiLerping) {
                transform.rotation = Quaternion.Slerp(oldRotation, newRotation, lerpCount);
            } else {
                transform.rotation = newRotation;
            }
        }

        private float lastframe = 0;
        private void LogEvery1Sec(string message) {
            if (Time.time - lastframe > 1) {
                SuperController.LogMessage(message);
                lastframe = Time.time;
            }
        }

        private float breatheLerp = 0;
        private IEnumerator VrHintButtonBreathingUpdate() {
            Color currentColor = hintButton.buttonColor;
            Color breatheColor = new Color(0.7f, 0, 0.5f, 0.4f);

            hintButton.GetComponent<RectTransform>().pivot = new Vector2(0.5f, 0.5f);

            while (lowerFrequencyChecks) {
                if (breatheLerp >= 1) hintButtonColorSummand = -0.04f;
                if (breatheLerp <= 0) hintButtonColorSummand = 0.04f;
                breatheLerp += hintButtonColorSummand;
                hintButton.buttonColor = Color.Lerp(currentColor, breatheColor, breatheLerp);
                hintButton.transform.localRotation = Quaternion.Lerp(Quaternion.Euler(0, 0, -10), Quaternion.Euler(0, 0, 10), breatheLerp);
                yield return new WaitForSeconds(0.07f);
            }
        }

        private void UpdateUi() {
            if (isVrActive) {
                UpdateCanvasLocalScale();
                UpdateVrUiPositionAndVisibility();
            } else {
                UpdateDesktopUiPosition();
            }
        }

        void FixedUpdate() {
            try {
                lerpCount += 0.02f;
                if (lerpCount > 1f) {
                    isWorldSpaceUiLerping = false;
                    lerpCount = 1f;
                }
            } catch (Exception e) {
                SuperController.LogError("Fixed update exception: " + e);
            }
        }

        // Update is called with each rendered frame by Unity
        void Update() {
            try {
                UpdateUi();
            } catch (Exception e) {
                SuperController.LogError("Update exception: " + e);
            }
        }
        #endregion

        // OnDestroy is where you should put any cleanup
        // if you registered objects to supercontroller or atom, you should unregister them here
        void OnDestroy() {
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            mainWindowCamera.enabled = true;
            lowerFrequencyChecks = false;
            SuperController.singleton.StopCoroutine(CheckEvery100millis());

            if (windowCameraRenderers.Length > 0) {
                for (int i = 0; i < windowCameraRenderers.Length; i++) {
                    windowCameraRenderers[i].transform.SetParent(renderersOriginalParents[i]);
                    windowCameraRenderers[i].gameObject.SetActive(true);
                    windowCameraRenderers[i].transform.position = displayControl.transform.position;
                    windowCameraRenderers[i].transform.rotation = displayControl.transform.rotation;
                    windowCameraRenderers[i].gameObject.layer = MINI_CAMERA_NOT_CULLED_LAYER;
                }
                windowCameraRenderers[0].transform.rotation *= Quaternion.Euler(180, 0, 0);
                windowCameraRenderers[1].transform.rotation *= Quaternion.Euler(0, 180, 0);

                windowCameraRenderers[0].material.shader = unlitTextureShader;
            }
            cameraOnStorable.val = initCameraOnValue;

            camPrevTrigger.Remove();
            camNextTrigger.Remove();
            foreach (AwasUIButton awasButton in buttons) {
                awasButton.buttonTrigger.Remove();
                awasButton.buttonAltTrigger.Remove();
            }
            if (worldSpaceUiCanvas != null) {
                SuperController.singleton.RemoveCanvas(worldSpaceUiCanvas);
                Destroy(worldSpaceUiCanvas.gameObject);
            }
            if (worldSpaceVrHintCanvas != null) {
                SuperController.singleton.StopCoroutine(VrHintButtonBreathingUpdate());
                SuperController.singleton.RemoveCanvas(worldSpaceVrHintCanvas);
                Destroy(worldSpaceVrHintCanvas.gameObject);
            }

            SetCameraTextureRenderResolution(256);
        }

        #region Boring
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
            JSONClass jsonClass = base.GetJSON(includePhysical, includeAppearance, forceStore);
            try {
                needsStore = true;

                JSONArray buttonsList = new JSONArray();
                foreach (AwasUIButton button in buttons) {
                    buttonsList.Add("", button.GetJSON());
                }
                jsonClass["buttons"] = buttonsList;
                jsonClass["camPrevTrigger"] = camPrevTrigger.GetJSON();
                jsonClass["camNextTrigger"] = camNextTrigger.GetJSON();
                jsonClass["cameraWillRenderToDesktop"] = cameraWillRenderToDesktop.val.ToString();
                jsonClass["miniCameraRenderResolution"] = miniCameraRenderResolution.val.ToString();
                jsonClass["miniCameraDisplayRotation"] = miniCameraDisplayRotation.val.ToString();
                jsonClass["uiIsAlwaysOnTop"] = uiIsAlwaysOnTop.val.ToString();
                jsonClass["uiIsVisibleInMiniCameraDisplay"] = uiIsVisibleInMiniCameraDisplay.val.ToString();
            } catch (Exception e) {
                SuperController.LogError($"AwasUI GetJSON exception: {e}");
            }

            return jsonClass;
        }

        public override void RestoreFromJSON(JSONClass jsonClass, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
            base.RestoreFromJSON(jsonClass, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            LoadConfig(jsonClass);
        }

        private bool isLoadingConfig = false;
        private void LoadConfig(JSONClass jsonConfig) {
            if (isLoadingConfig) return;
            isLoadingConfig = true;

            try {
                foreach (AwasUIButton button in buttons) {
                    button.buttonTrigger.Remove();
                    button.buttonAltTrigger.Remove();
                }

                JSONArray buttonsList = jsonConfig["buttons"].AsArray;
                for (int i = 0; i < buttonsList.Count; ++i) {
                    JSONClass buttonJson = buttonsList[i].AsObject;
                    AwasUIButton button = buttons[i];
                    button.RestoreFromJSON(buttonJson, base.subScenePrefix, base.mergeRestore, true);
                }
                camPrevTrigger.RestoreFromJSON(jsonConfig, base.subScenePrefix, base.mergeRestore, true);
                camNextTrigger.RestoreFromJSON(jsonConfig, base.subScenePrefix, base.mergeRestore, true);
                if (jsonConfig.HasKey("cameraWillRenderToDesktop")) {
                    cameraWillRenderToDesktop.val = bool.Parse(jsonConfig["cameraWillRenderToDesktop"]);
                }
                if (jsonConfig.HasKey("miniCameraRenderResolution")) {
                    miniCameraRenderResolution.val = jsonConfig["miniCameraRenderResolution"];
                }
                if (jsonConfig.HasKey("miniCameraDisplayRotation")) {
                    miniCameraDisplayRotation.val = jsonConfig["miniCameraDisplayRotation"];
                }
                if (jsonConfig.HasKey("uiIsAlwaysOnTop")) {
                    uiIsAlwaysOnTop.val = bool.Parse(jsonConfig["uiIsAlwaysOnTop"]);
                }
                if (jsonConfig.HasKey("uiIsVisibleInMiniCameraDisplay")) {
                    uiIsVisibleInMiniCameraDisplay.val = bool.Parse(jsonConfig["uiIsVisibleInMiniCameraDisplay"]);
                }
                ApplyStoredConfig();
            
            } catch (Exception e) {
                SuperController.LogError($"AwasUI Load error: {e}");
            } finally {
                isLoadingConfig = false;
            }
        }

        private void OnAtomRename(string oldid, string newid) {
            foreach (AwasUIButton button in buttons) {
                button.buttonTrigger.SyncAtomNames();
                button.buttonAltTrigger.SyncAtomNames();
            }
        }
    }

    public class SimpleTriggerHandler : TriggerHandler {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static SimpleTriggerHandler Instance {
            get {
                if (myInstance == null) {
                    myInstance = new SimpleTriggerHandler();
                }
                return myInstance;
            }
        }

        public static void LoadAssets() {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal() {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", prefab => myTriggerActionsPrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", prefab => myTriggerActionMiniPrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", prefab => myTriggerActionDiscretePrefab = prefab)) yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", prefab => myTriggerActionTransitionPrefab = prefab)) yield return x;
            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign) {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            }
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            }
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null) {
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            }
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t) {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t) {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI() {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt) {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

    // Wrapper for easier handling of custom triggers.
    public class EventTrigger : Trigger {
        public string Name {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public MVRScript Owner {
            get; private set;
        }

        private string name;
        private bool myNeedInit = true;

        public EventTrigger(MVRScript owner, string name) {
            Name = name;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public void OpenPanelActionStart() {
            if (!SimpleTriggerHandler.Loaded) {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit) {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void RestoreFromJSON(JSONClass jsonClass, string subScenePrefix, bool isMerge, bool setMissingToDefault) {
            if (jsonClass.HasKey(Name)) {
                JSONClass triggerJson = jsonClass[Name].AsObject;
                if (triggerJson != null) {
                    base.RestoreFromJSON(triggerJson, subScenePrefix, isMerge);
                }
            } else if (setMissingToDefault) {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger() {
            active = true;
            active = false;
        }

        public void TriggerTransition(float val) {
            this.transitionInterpValue = val;
        }
    }
    #endregion
}
